/*
 *   Octeon Watchdog driver
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2007 Cavium Networks
 */
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/ip.h>
#include <linux/string.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include "hal.h"
#include <linux/compat.h>
#include <asm/ioctl.h>       

struct timer_list watchdog_timers[2];
int timeout = 0;		/* A zero value is replaced with the max
				   timeout */
module_param(timeout, int, 0444);
MODULE_PARM_DESC(timeout,
		 " Watchdog timeout in milliseconds. First timeout causes a poke, \n"
		 "\t\tthe second causes an NMI, and the third causes a soft reset.");


/* master number  for watchdog device register */                             
#define WATCHDOG_MNO 64

/* 64 bits */
typedef struct WATCHDOG_CTRL_DATA {	
	uint32_t flag; // 
	uint16_t mode; // 0:off  1:on  2:blink
	uint16_t feq; // count of 1 sec
} WATCHDOG_CTRL_Data;

/* magic number for SCMD IOCTL operations */                             
#define SWMD_MAGIC 'B'

#define WATCHDOG_GET_MODE  _IOWR(SWMD_MAGIC, 21 , WATCHDOG_CTRL_Data) //set driver data     
#define WATCHDOG_SET_MODE  _IOW(SWMD_MAGIC, 22 , WATCHDOG_CTRL_Data) //set driver data     
#define WATCHDOG_SET_ON  _IOW(SWMD_MAGIC, 23 , WATCHDOG_CTRL_Data) //set driver data     
#define WATCHDOG_SET_OFF  _IOW(SWMD_MAGIC, 24 , WATCHDOG_CTRL_Data) //set driver data     


static int watchdog_ioctl(struct inode * inode, struct file * file, unsigned int cmd, unsigned long arg);     
#ifdef CONFIG_COMPAT
static long compat_watchdog_ioctl(struct file *filep, unsigned int cmd,  unsigned long arg);
#endif
static int watchdog_open( struct inode *inode, struct file *filp );

static struct file_operations watchdog_fops = {
	.ioctl =     watchdog_ioctl,
#ifdef CONFIG_COMPAT
	.compat_ioctl =		compat_watchdog_ioctl,
#endif
	.open =     watchdog_open,
};


static volatile unsigned char watchdog_flag =1;

static volatile int watchdogOpened = 0;  /* boolean value to track open status */

/*************************************************************************** 
* Function Name: led_open 
* Description  : Called when an application opens this device. 
* Returns      : 0 - success 
***************************************************************************/
static int watchdog_open( struct inode *inode, struct file *filp ) {
	if ( watchdogOpened == 0 )   {
		watchdogOpened = 1;   
	}
	return( 0 );
} /* endpoint_open */



/**
 * Poke the watchdog when an interrupt is received
 *
 * @param cpl
 * @param dev_id
 * @param regs
 *
 * @return
 */
 /*
static irqreturn_t watchdog_poke_irq(int cpl, void *dev_id)
{
	// We're alive, poke the watchdog 
	//printk("watchdog_poke_irq: Poking watchdog on core %d\n", cvmx_get_core_num());
	cvmx_write_csr(CVMX_CIU_PP_POKEX(cvmx_get_core_num()), 1);
	return IRQ_HANDLED;
}
*/

static void watchdog_incream_func(unsigned long ptr)
{
	/* We're alive, poke the watchdog */
	//printk("watchdog_incream_func: Poking watchdog on core %d\n", cvmx_get_core_num());
	if ( watchdog_flag ) {
		cvmx_write_csr(CVMX_CIU_PP_POKEX(cvmx_get_core_num()), 1);
	}
		watchdog_timers[0].expires = jiffies + HZ/2;
		add_timer(&watchdog_timers[0]);
	
}


/**
 * In the NMI handler we assume everything is broken. This
 * routine is used to send a raw byte out the serial port
 * with as minimal dependencies as possible.
 *
 * @param uart_index Uart to write to (0 or 1)
 * @param ch         Byte to write
 */
static inline void uart_write_byte(int uart_index, uint8_t ch)
{
	cvmx_uart_lsr_t lsrval;

	/* Spin until there is room */
	do {
		lsrval.u64 = cvmx_read_csr(CVMX_MIO_UARTX_LSR(uart_index));
	}
	while (lsrval.s.thre == 0);

	/* Write the byte */
	cvmx_write_csr(CVMX_MIO_UARTX_THR(uart_index), ch);
}


/**
 * Write a string to the uart
 *
 * @param uart_index Uart to use (0 or 1)
 * @param str        String to write
 */
static void uart_write_string(int uart_index, const char *str)
{
	/* Just loop writing one byte at a time */
	while (*str)
		uart_write_byte(uart_index, *str++);
}


/**
 * Write a hex number out of the uart
 *
 * @param uart_index Uart to use (0 or 1)
 * @param value      Number to display
 * @param digits     Number of digits to print (1 to 16)
 */
static void uart_write_hex(int uart_index, uint64_t value, int digits)
{
	int d;
	int v;
	for (d = 0; d < digits; d++) {
		v = (value >> ((digits - d - 1) * 4)) & 0xf;
		if (v >= 10)
			uart_write_byte(0, 'a' + v - 10);
		else
			uart_write_byte(0, '0' + v);
	}
}


/**
 * NMI stage 3 handler. NMIs are handled in the following manner:
 * 1) The first NMI handler enables CVMSEG and transfers from
 * the bootbus region into normal memory. It is careful to not
 * destroy any registers.
 * 2) The second stage handler uses CVMSEG to save the registers
 * and create a stack for C code. It then calls the third level
 * handler with one argument, a pointer to the register values.
 * 3) The third, and final, level handler is the following C
 * function that prints out some useful infomration.
 *
 * @param reg    Pointer to register state before the NMI
 */
void octeon_watchdog_nmi_stage3(uint64_t reg[32])
{
	const char *reg_name[] =
		{ "$0", "at", "v0", "v1", "a0", "a1", "a2", "a3",
		"a4", "a5", "a6", "a7", "t0", "t1", "t2", "t3",
		"s0", "s1", "s2", "s3", "s4", "s5", "s6", "s7",
		"t8", "t9", "k0", "k1", "gp", "sp", "s8", "ra"
	};
	uint64_t i;
	/* Save status and cause early to get them before any changes might
	   happen */
	uint64_t cp0_cause = read_c0_cause();
	uint64_t cp0_status = read_c0_status();

	/* Delay so all cores output is jumbled */
	cvmx_wait(100000000ull * cvmx_get_core_num());

	uart_write_string(0, "\r\n*** NMI Watchdog interrupt on Core 0x");
	uart_write_hex(0, cvmx_get_core_num(), 1);
	uart_write_string(0, " ***\r\n");
	for (i = 0; i < 32; i++) {
		uart_write_string(0, "\t");
		uart_write_string(0, reg_name[i]);
		uart_write_string(0, "\t0x");
		uart_write_hex(0, reg[i], 16);
		if (i & 1)
			uart_write_string(0, "\r\n");
	}
	uart_write_string(0, "\tepc\t0x");
	uart_write_hex(0, read_c0_epc(), 16);
	uart_write_string(0, "\r\n");

	uart_write_string(0, "\tstatus\t0x");
	uart_write_hex(0, cp0_status, 16);
	uart_write_string(0, "\tcause\t0x");
	uart_write_hex(0, cp0_cause, 16);
	uart_write_string(0, "\r\n");

	uart_write_string(0, "\tsum0\t0x");
	uart_write_hex(0,
		       cvmx_read_csr(CVMX_CIU_INTX_SUM0
				     (cvmx_get_core_num() * 2)), 16);
	uart_write_string(0, "\ten0\t0x");
	uart_write_hex(0,
		       cvmx_read_csr(CVMX_CIU_INTX_EN0
				     (cvmx_get_core_num() * 2)), 16);
	uart_write_string(0, "\r\n");
#if 0
	uart_write_string(0, "Code around epc\r\n");
	clear_c0_status(0x00400004);
	for (i = read_c0_epc() - 16; i < read_c0_epc() + 20; i += 4) {
		uart_write_string(0, "\t0x");
		uart_write_hex(0, i, 16);
		uart_write_string(0, "\t");
		uart_write_hex(0, *(uint32_t *) i, 16);
		uart_write_string(0, "\r\n");
	}
#endif
#if 0
	uart_write_string(0, "TLB\r\n");
	for (i = 0; i < 64; i++) {
		write_c0_index(i);
		asm volatile ("tlbr");
		uart_write_string(0, "\tHi 0x");
		uart_write_hex(0, read_c0_entryhi(), 16);
		uart_write_string(0, "\tPageMask 0x");
		uart_write_hex(0, read_c0_pagemask(), 16);
		uart_write_string(0, "\tLo0 0x");
		uart_write_hex(0, read_c0_entrylo0(), 16);
		uart_write_string(0, "\tLo1 0x");
		uart_write_hex(0, read_c0_entrylo1(), 16);
		uart_write_string(0, "\r\n");
	}
#endif
	uart_write_string(0, "*** Chip soft reset soon ***\r\n");
}


static void setup_interrupt(void *arg)
{
	cvmx_ciu_wdogx_t ciu_wdog;
	int core = cvmx_get_core_num();
	unsigned long threshold = (unsigned long) arg;

	/*
	request_irq(OCTEON_IRQ_WDOG0 + core, watchdog_poke_irq, SA_SHIRQ,
		    "watchdog", watchdog_poke_irq);
	*/

	/* Poke the watchdog to clear out its state */
	cvmx_write_csr(CVMX_CIU_PP_POKEX(core), 1);

	/* Finally enable the watchdog now that all handlers are installed */
	ciu_wdog.u64 = 0;
	ciu_wdog.s.len = threshold;
	ciu_wdog.s.mode = 3;	/* 3 = Interrupt + NMI + Soft-Reset */
	cvmx_write_csr(CVMX_CIU_WDOGX(core), ciu_wdog.u64);
}

/**************** Start the IO ctrl part **************************/
static int watchdog_ioctl(struct inode * inode, struct file * file,
                       unsigned int cmd, unsigned long arg)     
{                                                             
 	//LED_MODE_Data mode_data;
	
	switch (cmd) {
		case WATCHDOG_GET_MODE:
			{
 				//watchdog_flag = 0;
 				//printk(KERN_INFO "IOCTRL-WATCHDOG_GET_MODE: \n");
				return 3;
			
			}
		case WATCHDOG_SET_MODE:
			{
				cvmx_write_csr(CVMX_CIU_PP_POKEX(cvmx_get_core_num()), 1);
				//printk("IOCTRL-WATCHDOG_SET_MODE: Poking watchdog on core %d\n", cvmx_get_core_num());
				return 4;			
			}
		case WATCHDOG_SET_ON:
			{
 				watchdog_flag = 0;
 				//printk(KERN_INFO "IOCTRL-WATCHDOG_SET_ON: stop watchdog increas\n");
 				return 3;		
			}
		case WATCHDOG_SET_OFF:
			{
 				watchdog_flag = 1;
 				//printk(KERN_INFO "IOCTRL-WATCHDOG_SET_OFF: start watchdog increas\n");
 				return 3;		
			}
		default:
			printk(KERN_INFO "Unknown Command [%d]\n", cmd);
			break;
	}
	return 0 ;

}

#ifdef CONFIG_COMPAT
static long compat_watchdog_ioctl(struct file *filep, unsigned int cmd,
                              unsigned long arg)
{
	return watchdog_ioctl(filep->f_path.dentry->d_inode, filep, cmd, arg);
}
#endif




/**
 * Module/ driver initialization.
 *
 * @return Zero on success
 */
static int __init watchdog_init(void)
{
	uint64_t threshold;
	int i;
	extern void octeon_watchdog_nmi(void);

	//printk("Joe watchdog test. module init.\n");
	int retvalue;
	printk(KERN_INFO "Loading watchdog driver\n");

	retvalue = register_chrdev( WATCHDOG_MNO, "watchdog", &watchdog_fops );
	
	if (retvalue)
	{
		printk(KERN_WARNING "Initialization failed (/dev/watchdog)\n");
		return -1;
	}

	/* Watchdog time expiration length = The 16 bits of LEN represent the
	   most significant bits of a 24 bit decrementer that decrements every
	   256 cycles. */
	if (timeout)
		threshold = timeout * octeon_get_clock_rate() / 1000 >> (8 + 8);
	else {
		threshold = 65535;
		timeout =
			(65535ull << (8 + 8)) * 1000 / octeon_get_clock_rate();
	}
	if ((threshold < 10) || (threshold >= 65536)) {
		printk("Illegal watchdog timeout. Timeout must be between %u and %u.\n", (int) ((10ull << (8 + 8)) * 1000 / octeon_get_clock_rate()), (int) ((65535ull << (8 + 8)) * 1000 / octeon_get_clock_rate()));
		return -1;
	}

	/* Install the NMI handler */
	
	for (i = 0; i < 16; i++) {
		uint64_t *ptr = (uint64_t *) octeon_watchdog_nmi;
		cvmx_write_csr(CVMX_MIO_BOOT_LOC_ADR, i * 8);
		cvmx_write_csr(CVMX_MIO_BOOT_LOC_DAT, ptr[i]);
	}
	cvmx_write_csr(CVMX_MIO_BOOT_LOC_CFGX(0), 0x81fc0000);

	on_each_cpu(setup_interrupt, (void *) threshold, 1, 0);
	printk("Octeon watchdog driver loaded with a timeout of %d ms.\n",
	       timeout);
	
	init_timer(&watchdog_timers[0]);
	watchdog_timers[0].function = watchdog_incream_func;
	watchdog_timers[0].expires = jiffies + HZ/2;
	add_timer(&watchdog_timers[0]);
	
	return 0;
}


/**
 * Module / driver shutdown
 */
static void __exit watchdog_cleanup(void)
{
	int cpu;
	
	unregister_chrdev( WATCHDOG_MNO, "watchdog" );	
	//printk("Joe watchdog test. module clean.\n");
	preempt_disable();
	for (cpu = 0; cpu < NR_CPUS; cpu++) {
		if (cpu_online(cpu)) {
#ifdef CONFIG_SMP
			int core = cpu_logical_map(cpu);
#else
			int core = cvmx_get_core_num();
#endif
			/* Disable the watchdog */
			cvmx_write_csr(CVMX_CIU_WDOGX(core), 0);
			/* Free the interrupt handler */
			//free_irq(OCTEON_IRQ_WDOG0 + core, watchdog_poke_irq);
		}
	}
	preempt_enable();

	del_timer(&watchdog_timers[0]);

	printk (KERN_INFO "Unloading watchdog driver\n");

}

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Cavium Networks <support@caviumnetworks.com>");
MODULE_DESCRIPTION("Cavium Networks Octeon Watchdog driver.");
module_init(watchdog_init);
module_exit(watchdog_cleanup);
